/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EjbJarAnnotationMetadata.java 941 2006-07-26 09:06:13Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployment.annotations.metadata;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.ejb.ApplicationException;

import org.objectweb.easybeans.deployment.annotations.impl.JApplicationException;
import org.objectweb.easybeans.deployment.xml.struct.EJB3;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * This class represents the annotation metadata of all classes of an EjbJar file.
 * From this class, we can get metadata of all beans.
 * @author Florent Benoit
 */
public class EjbJarAnnotationMetadata {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(EjbJarAnnotationMetadata.class);

    /**
     * List of class annotations metadata.
     */
    private Map<String, ClassAnnotationMetadata> classesAnnotationMetadata = null;

    /**
     * Link to the Deployment Descriptor object.
     */
    private EJB3 ejb3 = null;

    /**
     * List of application exceptions used on this ejb-jar.
     */
    private Map<String, ApplicationException> applicationExceptions = null;

    /**
     * Constructor.
     */
    public EjbJarAnnotationMetadata() {
        classesAnnotationMetadata = new HashMap<String, ClassAnnotationMetadata>();
    }

    /**
     * Add annotation metadata for a given class.
     * @param classAnnotationMetadata annotation metadata of a class.
     */
    public void addClassAnnotationMetadata(final ClassAnnotationMetadata classAnnotationMetadata) {
        String key = classAnnotationMetadata.getClassName();
        // already exists ?
        if (classesAnnotationMetadata.containsKey(key)) {
            String msg = logger.getI18n().getMessage("EjbJarAnnotationMetadata.addClassAnnotationMetadata.alreadyPresent", key);
            logger.debug(msg);
            throw new IllegalStateException(msg);
        }
        classesAnnotationMetadata.put(key, classAnnotationMetadata);
    }


    /**
     * Get class annotation metadata.
     * @param className key of the map of annotations bean.
     * @return Bean annotation metadata of a given name.
     */
    public ClassAnnotationMetadata getClassAnnotationMetadata(final String className) {
        return classesAnnotationMetadata.get(className);
    }

    /**
     * Get collections of bean annotation metadata.
     * @return collections of bean annotation metadata.
     */
    public Collection<ClassAnnotationMetadata> getClassAnnotationMetadataCollection() {
        return classesAnnotationMetadata.values();
    }

    /**
     * @return the ejb3 deployment descriptor object.
     */
    public EJB3 getEjb3() {
        return ejb3;
    }

    /**
     * Sets the ejb3 deployment descriptor object.
     * @param ejb3 the ejb3 deployment descriptor object.
     */
    public void setEjb3(final EJB3 ejb3) {
        this.ejb3 = ejb3;
    }

    /**
     * Gets the list of application exceptions defined on this ejb jar metadata.
     * @return the list of application exceptions defined on this ejb jar metadata.
     */
    public Map<String, ApplicationException> getApplicationExceptions() {
        if (applicationExceptions != null) {
            return applicationExceptions;
        }

        // compute it
        applicationExceptions = new HashMap<String, ApplicationException>();

        // For each class, look if it is an application exception
        for (ClassAnnotationMetadata classMetadata : getClassAnnotationMetadataCollection()) {
            ApplicationException appException = classMetadata.getApplicationException();
            // found it then add it in the map.
            if (appException != null) {
                applicationExceptions.put(classMetadata.getClassName().replaceAll("/", "."), appException);
            }
        }

        // Add a default application exception (for checked exception)
        applicationExceptions.put("DEFAULT", new JApplicationException());

        return applicationExceptions;
    }

}
