/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JPool.java 940 2006-07-26 09:05:27Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.pool;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.easybeans.api.pool.Pool;
import org.objectweb.easybeans.api.pool.PoolAttributes;
import org.objectweb.easybeans.api.pool.PoolException;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;


/**
 * Abstract pool.<br>
 * Need to be extended to set-up the correct generic types used for BeanType and
 * Hint.<br>
 * @param <InstanceType> the type of the object that are managed by the pool
 *        (could be EasyBeansSLSB, etc.)
 * @param <Clue> a clue to retrieve a specific instance in the pool
 * @author Florent Benoit
 */
public class JPool<InstanceType, Clue> implements Pool<InstanceType, Clue>, PoolAttributes {

    /**
     * Default timeout (10s) in ms.
     */
    private static final long DEFAULT_TIMEOUT = 10000;

    /**
     * Default maximum of waiters (1000).
     */
    private static final int DEFAULT_MAX_WAITERS = 1000;

    /**
     * High Value for no limit for the connection pool.
     */
    private static final int NO_LIMIT = -1;

    /**
     * Logger used.
     */
    private JLog logger = JLogFactory.getLog(JPool.class);

    /**
     * Factory used for delegate creating, matching, removing actions.
     */
    private PoolFactory<InstanceType, Clue> poolFactory = null;

    /**
     * List of available objects in the pool.
     */
    private List<InstanceType> availableList = null;

    /**
     * List of objects which are currently busy and not available from the pool.
     */
    private List<InstanceType> busyList = null;

    /**
     * Information on entries managed in the pool.
     */
    private Map<InstanceType, PoolEntryStatistics> infoMap = null;

    /**
     * Minimum size of the pool.
     */
    private int minSize = 0;

    /**
     * Maximum size of the pool.
     */
    private int maxSize = NO_LIMIT;

    /**
     * Initial size of the pool.
     */
    private int initSize = 0;

    /**
     * Maximum size of the pool is strict.
     */
    private boolean strict = true;

    /**
     * Max nb of waiters allowed to wait for an instance.
     */
    private int maxWaiters = DEFAULT_MAX_WAITERS;

    /**
     * Max nb of milliseconds to wait for an object when the pool is full.
     */
    private long waiterTimeout = DEFAULT_TIMEOUT;

    /**
     * Current number of waiters.
     */
    private int currentWaiters = 0;

    /**
     * Maximum number of waiters during current period.
     */
    private int highMaxWaiters = 0;

    /**
     * Total nb of waiters since pool creation.
     */
    private int totalWaiters = 0;

    /**
     * total waiting time in milliseconds.
     */
    private long totalWaitingTime = 0;

    /**
     * Maximum waiting time during current period.
     */
    private long highWaitingTime = 0;

    /**
     * Total number of instances since the pool creation.
     */
    private int servedInstance = 0;

    /**
     * Total nb of instance not served due to timeout.
     */
    private int rejectedTimeout = 0;

    /**
     * Total nb of instance not served due to an overflow of waiters.
     */
    private int rejectedFull = 0;

    /**
     * Builds a new pool.
     * @param poolFactory factory used for delegating create, remove, etc.
     */
    public JPool(final PoolFactory<InstanceType, Clue> poolFactory) {

        // factory use to delegate some operations
        this.poolFactory = poolFactory;

        // available entries
        this.availableList = new ArrayList<InstanceType>();

        // entries send to client, which are busy
        this.busyList = new ArrayList<InstanceType>();

        // statistics on entries
        infoMap = new HashMap<InstanceType, PoolEntryStatistics>();
    }

    /**
     * Gets an object from the pool.
     * @return an instance of an object with type Type.
     * @throws PoolException if instance cannot be returned.
     */
    public InstanceType get() throws PoolException {
        // delegates to getter method with a parameter
        return this.get(null);
    }

    /**
     * Gets an object by using a specific hint.
     * @param clue attribute used to retrieve a given instance
     * @return a specific instance of the resource.
     * @throws PoolException if instance cannot be returned.
     */
    public synchronized InstanceType get(final Clue clue) throws PoolException {
        InstanceType instance = null;

        // time to wait when pool is full
        long timeToWait = waiterTimeout;

        // Try to find an instance
        // Enter in the big loop :)
        while (instance == null) {

            // instances are available
            if (!availableList.isEmpty()) {
                logger.debug("Instances are available");

                int indexToExtract = -1;

                // No clue, get first instance available
                if (clue == null) {
                    logger.debug("No clue, return the first available entry.");
                    indexToExtract = 0;
                } else {
                    logger.debug("Clue is used, try to get a matching entry.");
                    // need to find a matching instance.
                    // use this kind of loop to call remove() method
                    // with an index instead of an element.
                    for (int i = 0; i < availableList.size(); i++) {
                        InstanceType tmpInstance = availableList.get(i);
                        if (poolFactory.isMatching(tmpInstance, clue)) {
                            logger.debug("Found a matching instance, instance = {0}", tmpInstance);
                            indexToExtract = i;
                            break;
                        }
                    }
                    // Note : instance could have not be found
                }
                // An instance is available ?
                if (indexToExtract != -1) {
                    logger
                            .debug("Remove from available instance, item with index {0}", Integer
                                    .valueOf(indexToExtract));
                    instance = availableList.remove(indexToExtract);
                    busyList.add(instance);
                }
            }

            /**
             * If here :
             * <ul>
             * <li> - No objects available in the pool</li>
             * <li>or - No matching instance in the pool</li>
             * </ul>
             */
            if (instance == null) {
                // current size of managed entries (available and busy)
                int curSize = availableList.size() + busyList.size();

                // limit not reached (or pool without limit)
                if (maxSize == NO_LIMIT || curSize < maxSize) {
                    logger.debug("Maximum size not reached, can create a new entry with clue = {0}", clue);
                    instance = poolFactory.create(clue);
                    busyList.add(instance);
                } else if (availableList.size() > 0) {
                    // limit reached but pool entries still available
                    logger.debug("limit reached but entries available in the pool. Replace one by a new one.");

                    // get first entry found (that will be replaced)
                    InstanceType previousInstance = availableList.get(0);

                    // remove it from the delegating factory.
                    // Factory is notified that the instance is being removed.
                    poolFactory.remove(previousInstance);

                    // now, remove the entry from pool
                    availableList.remove(previousInstance);
                    // and from statistics
                    infoMap.remove(previousInstance);

                    // As previous instance has been removed, new one can be
                    // created
                    instance = poolFactory.create(clue);
                    // this instance is now busy.
                    busyList.add(instance);
                } else {
                    logger.debug("limit reached and no available instances in the pool");
                    boolean timeoutWokenUp = true;
                    long startSleeping = 0;

                    // Need to wait
                    if (timeToWait > 0) {
                        if (currentWaiters < maxWaiters) {
                            currentWaiters++;
                            // Store the maximum concurrent waiters
                            if (highMaxWaiters < currentWaiters) {
                                highMaxWaiters = currentWaiters;
                            }
                            if (startSleeping == 0) {
                                startSleeping = System.currentTimeMillis();
                                logger.debug("Waiting an instance");
                            }
                            try {
                                wait(timeToWait);
                            } catch (InterruptedException ie) {
                                logger.warn("Waiter has been interrupted", ie);
                            } finally {
                                currentWaiters--;
                            }
                            long sleepTime = System.currentTimeMillis() - startSleeping;
                            // check if waiter has been woken up
                            timeoutWokenUp = (waiterTimeout - sleepTime <= 0);

                            // increment counters
                            totalWaiters++;
                            totalWaitingTime += sleepTime;
                            if (highWaitingTime < sleepTime) {
                                highWaitingTime = sleepTime;
                            }
                            // continue if we were only notified. (no timeout)
                            if (!timeoutWokenUp) {
                                continue;
                            }
                        }
                    }

                    // timeout expired and no entry is available
                    if (timeoutWokenUp && availableList.isEmpty() && busyList.size() >= maxSize) {
                        if (startSleeping > 0) {
                            rejectedTimeout++;
                            logger.warn("Cannot create an instance due to a timeout");
                        } else {
                            rejectedFull++;
                            logger.warn("Cannot create an instance");
                        }
                        throw new PoolException("No more instances available");
                    }
                }
            }
        }

        // TODO: delegate to factory the entry
        PoolEntryStatistics poolEntryStat = new PoolEntryStatistics();

        infoMap.put(instance, poolEntryStat);

        // printLists();

        // recomputeBusy();

        servedInstance++;
        return instance;
    }

    /**
     * Puts back the instance in the pool so it can be reused.
     * @param instance which will be put back in the pool.
     * @throws PoolException if instance is not released.
     */
    public synchronized void release(final InstanceType instance) throws PoolException {

        if (!busyList.contains(instance)) {
            logger.debug("Attempt to release inactive resource {0}. Probably discarded.", instance);
            return;
        }
        busyList.remove(instance);
        logger.debug("Put back into the pool the instance {0}", instance);

        availableList.add(instance);
        infoMap.remove(instance);

        // Notify 1 thread waiting for that an instance is available.
        if (currentWaiters > 0) {
            notify();
        }
    }

    /**
     * Discard the instance which is in the pool.
     * @param instance which will be discarded.
     * @throws PoolException if instance is not discarded
     */
    public synchronized void discard(final InstanceType instance) throws PoolException {
        if (!busyList.contains(instance)) {
            throw new PoolException("Attempt to discard an inactive resource(" + instance + ")");
        }

        // Notify the factory that it's going to be removed
        poolFactory.remove(instance);

        // Remove instance from the busy & info list
        busyList.remove(instance);
        infoMap.remove(instance);


        // Notify 1 thread waiting that an instance has been freed
        if (currentWaiters > 0) {
            notify();
        }

    }

    /**
     * Start the pool.<br>
     * It could create initial instances if specified.
     * @throws PoolException if initialization fails
     */
    public void start() throws PoolException {

    }

    /**
     * Stop this pool.
     * @throws PoolException if destroy fails
     */
    public void stop() throws PoolException {

    }

}
