/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EasyBeansContextListener.java 1026 2006-08-04 14:54:10Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.server.war;
import static org.objectweb.easybeans.util.url.URLUtils.fileToURL2;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.server.EasyBeans;
import org.objectweb.easybeans.server.Embedded;
import org.objectweb.easybeans.server.EmbeddedConfigurator;
import org.objectweb.easybeans.server.EmbeddedException;
import org.objectweb.easybeans.server.ServerConfig;
import org.objectweb.easybeans.util.url.URLUtilsException;

/**
 * Listener class which is notified of lifecycle events. It allows to start
 * EasyBeans at startup and stop it at the end.
 * @author Florent Benoit
 */
public class EasyBeansContextListener implements ServletContextListener {

    /**
     * Resource containing the list of files.
     */
    private static final String COMMON_LIBRARIES_LIST = "easybeans-exported-libraries.lst";

    /**
     * Default XML file.
     */
    public static final String DEFAULT_XML_FILE = "org/objectweb/easybeans/server/war/easybeans-default.xml";

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(EasyBeansContextListener.class);

    /**
     * Embedded instance.
     */
    private Embedded embedded = null;

    /**
     * List of available supported container.
     */
    private enum AvailableWebContainer {
        /**
         * Containers supported.
         */
        TOMCAT, JETTY, UNKNOWN
    }

    /**
     * Container detected ?.
     */
    private AvailableWebContainer containerType;

    /**
     * Notification that the web application initialization process is starting.
     * All ServletContextListeners are notified of context initialization before
     * any filter or servlet in the web application is initialized.
     * @param servletContextEvent event class for notifications about changes to
     *        the servlet context of a web application.
     */
    public void contextInitialized(final ServletContextEvent servletContextEvent) {

        // Create and configure EasyBeans embedded server
        if (embedded == null) {

            // user configuration ?
            URL xmlConfigurationURL = Thread.currentThread().getContextClassLoader().getResource(EasyBeans.USER_XML_FILE);

            if (xmlConfigurationURL == null) {
                logger.warn("No {0} resource found in classpath, use default settings", EasyBeans.USER_XML_FILE);
                xmlConfigurationURL = Thread.currentThread().getContextClassLoader().getResource(DEFAULT_XML_FILE);
            }
            try {
                embedded = EmbeddedConfigurator.create(xmlConfigurationURL);
            } catch (EmbeddedException e) {
                throw new IllegalStateException("Cannot create the embedded server", e);
            }

            // Detects Tomcat/Jetty/etc.
            String ejb3ParentDir = detectWebContainerDirectory();
            logger.info("Detecting '" + containerType + "' web container");

            // Create ejb3s directory
            File workDir = new File(ejb3ParentDir + File.separator + "ejb3s");
            workDir.mkdir();

            ServerConfig serverConfig = getServerConfig();
            serverConfig.setShouldWait(false);
            serverConfig.setUseMBeans(true);
            serverConfig.setUseNaming(false);
            logger.info("Set ejb3 path to ''{0}''", workDir);
            serverConfig.setEjb3Path(workDir.getAbsolutePath());

            // TODO ejb3 path ??

            registerLibraries(servletContextEvent);
            try {
                embedded.start();
            } catch (EmbeddedException e) {
                throw new IllegalStateException("Cannot start embedded EasyBeans server", e);
            }
        }
    }

    /**
     * Notification that the servlet context is about to be shut down. All
     * servlets and filters have been destroy()ed before any
     * ServletContextListeners are notified of context destruction.
     * @param servletContextEvent event class for notifications about changes to
     *        the servlet context of a web application.
     */
    public void contextDestroyed(final ServletContextEvent servletContextEvent) {
        if (embedded != null) {
            logger.info("Stopping EasyBeans embedded server...");
            try {
                embedded.stop();
            } catch (EmbeddedException e) {
                throw new IllegalStateException("Cannot stop the embedded server", e);
            } finally {
                embedded = null;
            }
        }
    }

    /**
     * Gets the directory for EJB3 depending on the servlet container.
     * @return path to catalina.base/jetty.home/etc.
     */
    private String detectWebContainerDirectory() {
        // catalina base ?
        String catalinaBase = System.getProperty("catalina.base");
        if (catalinaBase != null) {
            containerType = AvailableWebContainer.TOMCAT;
            return catalinaBase;
        }

        // Jetty ?
        String jettyHome = System.getProperty("jetty.home");
        if (jettyHome != null) {
            containerType = AvailableWebContainer.JETTY;
            return jettyHome;
        }

        // not found, return tmp directory
        containerType = AvailableWebContainer.UNKNOWN;
        return System.getProperty("java.io.tmpdir");
    }

    /**
     * Register the EasyBeans libraries in the root classloader (allowing each
     * applications to communicate with EasyBeans).
     * @param servletContextEvent event class for notifications about changes to
     *        the servlet context of a web application.
     */
    private void registerLibraries(final ServletContextEvent servletContextEvent) {
        if (containerType.equals(AvailableWebContainer.TOMCAT)) {
            registerTomcat(servletContextEvent);
        }

    }

    /**
     * Register the EasyBeans libraries for Tomcat.
     * @param servletContextEvent event class for notifications about changes to
     *        the servlet context of a web application.
     */
    private void registerTomcat(final ServletContextEvent servletContextEvent) {

        // Current Classloader
        ClassLoader currentCL = Thread.currentThread().getContextClassLoader();

        // Now, search shared classloader (child of server CL)
        // Assume that it is a correct Tomcat installation
        ClassLoader cl = currentCL.getParent().getParent();
        // Check class :
        if (!"org.apache.catalina.loader.StandardClassLoader".equals(cl.getClass().getName())) {
            throw new IllegalStateException("Didn't find common classloader");
        }

        // Cast classLoader in URL Classloader
        URLClassLoader urlCl = (URLClassLoader) cl;
        // Get addURL method
        Method addURL = null;
        try {
            addURL = URLClassLoader.class.getDeclaredMethod("addURL", new Class[] {URL.class});
        } catch (SecurityException e) {
            throw new IllegalArgumentException("Cannot get addURL method on the class '" + URLClassLoader.class + "'.", e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException("Cannot get addURL method on the class '" + URLClassLoader.class + "'.", e);
        }
        addURL.setAccessible(true);

        // Add libraries
        // / Get Carol jar
        ServletContext sContext = servletContextEvent.getServletContext();
        List<String> libraries = new ArrayList<String>();
        getLibraries(sContext, libraries);

        for (String library : libraries) {
            URL url = null;
            try {
                url = fileToURL2(new File(library));
            } catch (URLUtilsException e) {
                throw new IllegalArgumentException("Cannot get URL on file '" + library + "'.", e);
            }
            try {
                logger.debug("Registering '" + url + "'.");
                addURL.invoke(urlCl, url);
            } catch (IllegalArgumentException e) {
                throw new IllegalStateException("Cannot register the URL '" + url + "'.", e);
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Cannot register the URL '" + url + "'.", e);
            } catch (InvocationTargetException e) {
                throw new IllegalStateException("Cannot register the URL '" + url + "'.", e);
            }
        }
        addURL.setAccessible(false);

    }

    /**
     * Retrieve a List of libraries to be loaded by TopLevel ClassLoaders.
     * @param sContext ServletContext.
     * @param libraries List of libs.
     */
    protected void getLibraries(final ServletContext sContext, final List<String> libraries) {
        // Files that needs to be in the commons classloader are stored in a
        // properties file.

        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();

        URL url = classLoader.getResource(COMMON_LIBRARIES_LIST);
        if (url == null) {
            throw new IllegalStateException("No resource named '" + COMMON_LIBRARIES_LIST
                    + "' was found in the current classloader.");
        }

        // got url, open it
        URLConnection urlConnection = null;
        try {
            urlConnection = url.openConnection();
        } catch (IOException e) {
            throw new IllegalStateException("Cannot open connection on the URL '" + url + "'.", e);
        }
        urlConnection.setDefaultUseCaches(false);

        Reader reader = null;
        try {
            try {
                reader = new InputStreamReader(urlConnection.getInputStream());
            } catch (IOException e) {
                throw new IllegalStateException("Cannot get inputstream on the URL '" + url + "'.", e);
            }
            BufferedReader bufferedReader = new BufferedReader(reader);
            String fileName = null;
            try {
                while ((fileName = bufferedReader.readLine()) != null) {
                    fileName = fileName.trim();
                    // only lines with values
                    if (fileName.length() >= 0) {
                        // Get filename
                        libraries.add(sContext.getRealPath("/WEB-INF/lib/" + fileName.trim()));
                    }
                }
            } catch (IOException e) {
                throw new IllegalStateException("Cannot read the inputstream on the URL '" + url + "'.", e);
            } finally {
                if (bufferedReader != null) {
                    try {
                        bufferedReader.close();
                    } catch (IOException e) {
                        throw new IllegalStateException("Cannot close reader object on the URL '" + url + "'.", e);
                    }
                }
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException e) {
                    throw new IllegalStateException("Cannot close reader object on the URL '" + url + "'.", e);
                }
            }
        }
    }

    /**
     * @return returns the Easybeans ServerConfig object.
     */
    protected ServerConfig getServerConfig() {
        return embedded.getServerConfig();
    }
}
